@@ -1,8 +1,8 @@
 #!/usr/bin/env -S guile -e start -s
 !#
 ;;; coding: utf-8
-;;; $Id: debian-copyrighter.scm,v 1.50 2023/01/12 23:48:09 fpp Exp fpp $
-;;; Copyright © 2022 Frank Pursel <frank.pursel@gmail.com>
+;;; $Id: debian-copyrighter.scm,v 1.69 2023/12/14 16:31:57 fpp Exp fpp $
+;;; Copyright © 2022-2023 Frank Pursel <frank.pursel@gmail.com>
 
  ;; The debian-copyrighter is free software; you can redistribute it and/or modify it
  ;; under the terms of the GNU General Public License as published by
@@ -47,7 +47,10 @@
 
 (define (licensed-file-license->string rec)
   "license must be a symbol for processing but then a string at output."
-  (symbol->string (licensed-file-license rec)))
+  (let ((lfl (licensed-file-license rec)))
+    (if lfl
+	(symbol->string (licensed-file-license rec))
+	"No License Found")))
 
 (define-record-type <copyright>
   (make-copyright tag year-list)
@@ -55,6 +58,8 @@
   (tag copyright-tag set-copyright-tag!)
   (year-list copyright-year-list set-copyright-year-list!))
 
+(define diagnostic-str "Done!\n")
+
 ;;; Tools to compress a list of numeric years.
 (define (group-adj-aux sorted-l res)
   (if (null? sorted-l)
@@ -127,7 +132,15 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
 
 (define (expand-hyphen-yrs h-yr-str)
   (let* ((str-l (string-split h-yr-str (string-ref h-yr-str 4)))
-	 (n-lst (sort (map string->number str-l) <=))
+	 (fixed-2digit-str-l (map (lambda (ns)
+				    (if (= (string-length ns) 4)
+					ns
+					(let ((num (string->number ns)))
+					  (if (> 50 num)
+					      (format #f "~a" (+ 2000 num))
+					      (format #f "~a" (+ 1900 num))))))
+				  str-l))
+	 (n-lst (sort (map string->number fixed-2digit-str-l) <=))
 	 (low (car n-lst))
 	 (high (cadr n-lst)))
     (string-join (map number->string (iota (- high low -1) low)) ",")
@@ -135,7 +148,8 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
   )
 
 
-(define hy-yrs-cre (make-regexp "(19|20)[8,9,0,1,2][0-9][-–](19|20)[8,9,0,1,2][0-9]"))
+;;(define hy-yrs-cre-old (make-regexp "(19|20)[8,9,0,1,2][0-9][-–](19|20)[8,9,0,1,2][0-9]"))
+(define hy-yrs-cre (make-regexp "(19|20)[8,9,0,1,2][0-9][-–][0-9][0-9]+"))
 
 (define (dehyphenate-line str)
   (let ((match? (regexp-exec hy-yrs-cre str)))
@@ -154,9 +168,21 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
 ;(test-end "Line dehyphenation")
 
 
+(define (canon-year n)
+  (if (< n 100)
+      (if (> n 50)
+	  (+ n 1900)
+	  (+ n 2000))
+      n))
+
 (define (years-str->nlst str)
-  (let ((dehyphenated (dehyphenate-line str)))
-    (filter number? (map string->number (map string-trim-both (string-split dehyphenated #\,)))))
+  (let ((dehyphenated (dehyphenate-line
+		       str)))
+    (map canon-year
+	 (filter number?
+		 (map string->number
+		      (map string-trim-both (string-split dehyphenated #\,)))))
+    )
   )
 
 ;; (test-begin "strings-to-lists-of-numbers")
@@ -244,16 +270,31 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
 ;;   (test-assert (not ((test-lic? (list 'x-sym "A" "tree")) "A buffalo")))
 ;;   (test-assert (eq? 'xsym ((test-lic? (list 'xsym "A" "tree")) "A tree and more"))))
 
-(define years-re (make-regexp "((20|19)[8,9,0,1,2][0-9][-–,]? ?)+"))
+;(define years-re (make-regexp "(20|19)[8,9,0,1,2][0-9][ \t]?[-–,]?[ \t]?[0-9][0-9]+"))
+(define years-re (make-regexp "[-–,0-9][-–,0-9[:space:]]+"))
 
 
 (define (make-copyright-from match)
-  (let* ((years-mat (regexp-exec years-re (match:suffix match)))
-	 (years-str (if years-mat (match:substring years-mat) (match:substring match)))
-	 (tag-str (if years-mat
-		      (string-drop (match:suffix match) (string-length years-str))
-		      (match:substring match))))
-    (make-copyright tag-str (years-str->nlst years-str))))
+  (let* (
+	 (copyright-info (match:substring match 3))
+	 (years-mat (regexp-exec years-re copyright-info))
+	 (years-str (if years-mat
+			(match:substring years-mat)
+					; (match:substring match)
+			""
+			))
+	 (tag-str (fold
+		   (lambda (c s) (string-trim-both s c))
+		   (if years-mat
+		       (if (equal? "" (match:prefix years-mat))
+			   (string-drop copyright-info (string-length years-str))
+			   (match:prefix years-mat))
+		       copyright-info)
+		   ;; Note that order and repetitions matter here.
+		   (list #\space #\tab #\: #\. #\- #\* #\* #\\ #\/ #\; #\% #\space #\tab))))
+    (if (> 0 (string-length tag-str))
+	"Empty Copyright Tag"
+	(make-copyright tag-str (years-str->nlst years-str)))))
   
 
 (define copyr-re (make-regexp
@@ -263,12 +304,19 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
 			 ") "))
 		  regexp/icase))
 
+;; (define copyr-re-i (make-regexp
+;; 		  (string-concatenate
+;; 		   (list "copyright( \\([Cc]\\)| "
+;; 			 (list->string (list #\x00A9))
+;; 			 ")* "))
+;; 		  regexp/icase))
+
 (define copyr-re-i (make-regexp
-		  (string-concatenate
-		   (list "copyright( \\([Cc]\\)| "
-			 (list->string (list #\x00A9))
-			 ")* "))
-		  regexp/icase))
+		    (string-concatenate
+		     (list "(copyright|!copyright notice|!copyright HOLD)[[:space:]0-9]+(\\([Cc]\\)|"
+			   (list->string (list #\x00A9))
+			   ")[[:space:]]*(.*)"))
+		    regexp/icase))
 
 (define (copyrights-of-aux lof)
   (if (null? lof)
@@ -280,6 +328,13 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
 	     (copyrights-of-aux (cdr lof)))
 	    (copyrights-of-aux (cdr lof))))))
 
+;;; Enabling additional testing
+(define (copyright-from-str str)
+  (let ((m? (regexp-exec copyr-re-i str)))
+    (if m?
+	(make-copyright-from m?)
+	#f)))
+
 (define copyright-po-cre (make-regexp "([-–,] ?(20|19)[8,9,1,2][0-9])+\\.?[[:blank:]]*$"))
 
 (define (make-copyright-from-po match)
@@ -323,7 +378,15 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
 				   name
 				   (dir-from-full-file-name name)
 				   (license-of lof license-tests)
-				   (copyrights-of lof (basename name))))
+				   ;;; Flagging copyright problems
+				   (let ((copyrights 
+					   (copyrights-of lof (basename name))))
+				     (if (every copyright? copyrights)
+					 copyrights
+					 (begin
+					   (pk name)
+					   (filter copyright? copyrights))))
+				   ))
 	      )
 	  ))
     res
@@ -339,6 +402,65 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
     (filter empty-copyr? copyr-list))
   )
 
+
+;;; Function to do things with transforms that are defined below.
+(define (str-transfm transform)
+  (lambda (str)
+    (let* ((cre (car transform))
+	   (subst-str (cadr transform))
+	   (m? (regexp-exec cre str))
+	   )
+      (if m?
+	  (regexp-substitute
+	   #f
+	   m?
+	   'pre
+	   subst-str
+	   'post)
+	  str))))
+
+(define (gbl-str-transfm transform)
+  (lambda (str)
+    (let* ((cre (car transform))
+	   (subst-str (cadr transform))
+	   )
+      (regexp-substitute/global
+       #f
+       cre
+       str
+       'pre
+       subst-str
+       'post)
+      )))
+
+;;; Defining some simple string transformations.  They define regular expressions that
+;;; when matched are replaced by a string constant.
+;;;
+;;; They are simple and composable.
+(define non-lic-chars (list
+		       (make-regexp "[|_]+" regexp/extended) ""))
+(define invalid-lic-1 (list
+		       (make-regexp "GPL-2\\.0-or-later" regexp/basic) "GPL-2.0+"))
+(define invalid-lic-2 (list (make-regexp "GPL-2\\.0-only" regexp/basic) "GPL-2.0"))
+(define invalid-lic-3 (list
+		       (make-regexp "GPL-3\\.0-or-later" regexp/basic) "GPL-3.0+"))
+(define invalid-lic-4 (list
+		       (make-regexp "GPL-3\\.0-only" regexp/basic) "GPL-3.0"))
+(define GPL-space-to-hyphen (list (make-regexp "GPL 2" regexp/basic) "GPL-2"))
+(define or-later (list (make-regexp "-or-later ") "+ "))
+
+;; Here we compose multiple transforms into a single operation.
+(define spdx-license-canonicalize
+  (compose
+   (gbl-str-transfm non-lic-chars)
+   (str-transfm invalid-lic-1)
+   (str-transfm invalid-lic-2)
+   (str-transfm invalid-lic-3)
+   (str-transfm invalid-lic-4)
+   (str-transfm GPL-space-to-hyphen)
+   (str-transfm or-later)
+   ))
+
 (define (spdx-license? lof)
   (spdx-license-auxx lof #f 15))
 
@@ -354,7 +476,7 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
 (define prune-end-comment-re
   (make-regexp "(.*)[[:space:]]*\\*/"))
 
-(define (spdx-license-identifier-value line)
+(define (spdx-license-identifier-value-raw line)
   (let ((value+end
 	 (cond
 	  ((regexp-exec spdx-lic-cparen-re line) => (cut match:substring <> 1))
@@ -363,15 +485,24 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
 	  (else #f))))
     (if value+end
 	(let ((end-comment-match? (regexp-exec prune-end-comment-re value+end)))
-	  (string->symbol
-	   (string-trim-right (if end-comment-match?
-				  (match:substring end-comment-match? 1)
-				  value+end))))
+	  (string-trim-right (if end-comment-match?
+				 (match:substring end-comment-match? 1)
+				 value+end)))
 	#f
 	)
     )
   )
 
+(define (spdx-license-identifier-value line)
+  (let ((raw-license-identifier (spdx-license-identifier-value-raw line)))
+    (if (string? raw-license-identifier)
+	(string->symbol
+	 (spdx-license-canonicalize
+	  raw-license-identifier))
+	raw-license-identifier)
+    )
+  )
+
 (define (spdx-license-auxx lof res lines-to-check)
   (if (or (null? lof) res (< lines-to-check 1))
       res
@@ -421,7 +552,7 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
 				   (trim-base name)
 				   (dir-from-full-file-name name)
 				   (licensing? text-lic? spdx-lic?)
-				   (filter-empty-copyrights
+				   (filter copyright?
 				    (copyrights-of lof (basename name)))))
 		  
 	      ))
@@ -659,7 +790,7 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
 (set-record-type-printer!
  <copyright>
  (lambda (rec port)
-   (format port " ~a ~a~%"
+   (format port "~a ~a~%"
 	   (display-cprt-yrs (copyright-year-list rec))
 	   (copyright-tag rec))))
 (set-record-type-printer!
@@ -671,13 +802,19 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
 		 "File: ~a~%Copyright:\n~a~%License: ~a~%"
 		 name
 		 (string-join
-		  (map (lambda (copyr-rec) (string-concatenate
-					    (list
-					     " "
-					     (display-cprt-yrs
-					      (copyright-year-list copyr-rec))
-					     " "
-					     (copyright-tag copyr-rec))))
+		  (map (lambda (copyr-rec)
+			 (format #f "~a ~a"
+				 (display-cprt-yrs
+				  (copyright-year-list copyr-rec))
+				 (copyright-tag copyr-rec))
+			 ;; (string-concatenate
+			 ;;  (list
+			 ;;   " "
+			 ;;   (display-cprt-yrs
+			 ;;    (copyright-year-list copyr-rec))
+			 ;;   " "
+			 ;;   (copyright-tag copyr-rec)))
+			       )
 		       (licensed-file-copyright-list rec))
 		  "\n")
 		 (licensed-file-license->string rec))
@@ -691,7 +828,7 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
 			       (let ((copyr (cdr ae)))
 				 (string-concatenate
 				  (list
-				   " "
+				   ;; " "
 				   (display-cprt-yrs (copyright-year-list copyr))
 				   " "
 				   (copyright-tag copyr)))))
@@ -711,6 +848,14 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
      sorted-al)
     ))
 
+;; (define (license-filename-sort d-hash licenses)
+;;   (sort licenses
+;; 	(lambda (x y)
+;; 	  (string<=?
+;; 	   (licensed-file-name (licensed-files/license d-hash x))
+;; 	   (licensed-file-name (licensed-files/license d-hash y))
+;; 		     ))))
+
 (define (default-output dir-tree exclusion-list)
   (let* ((d-hash (hash-for-dir-tree-w-exclusions dir-tree exclusion-list))
 	 (license-freqs (license-counts d-hash (licenses-found d-hash)))
@@ -725,10 +870,11 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
        (let ((licensed-file-rec (licensed-files/license d-hash lic)))
 	 (format #t "Files:\n ~a~%"
 		 (string-join
-		   (licensed-file-name licensed-file-rec) "\n "))
+		  (sort (licensed-file-name licensed-file-rec)
+			string<=?) "\n "))
 	 (display licensed-file-rec)
 	 (newline)))
-     rem-licenses)
+     rem-licenses )
     ))
 
 (define (margs opt-name parsed-opts default-regexp)
@@ -741,13 +887,15 @@ returns \"1995,2005\".  Or the empty string for no copyrights."
 
 
 (define (start args)
-  (let* ((option-spec '((dir (single-char #\d) (value #t))
+  (let* ((option-spec '((single-file (single-char #\f) (value #t))
+			(dir (single-char #\d) (value #t))
 			(exclusions (single-char #\x) (value #t))
 			(list-files (single-char #\L) (value #f))
 			(short-years (single-char #\s) (value #f))
 			(all (single-char #\A) (value #f))
 			(help (single-char #\h) (value #f))))
 	 (options (getopt-long args option-spec))
+	 (filename (option-ref options 'single-file #f))
 	 (dir-provided (option-ref options 'dir #f))
 	 (ex-list (margs 'exclusions options '()))
 	 (list-files (option-ref options 'list-files #f))
@@ -764,6 +912,8 @@ debian-copyrighter [options]
   -x,  --exclusions Possibly multiple regular-expressions for
                dirs and/or files to be excluded from analysis
   -A,  --all   Lists license, copyright holders for all files (long)
+  -f,  --single-file (pre-empts -d, -x, and -A) Analyze the given 
+               filename (for testing).
 
 Scans the provided dir for files containing license
 information and outputs collective summary of copyright holder
@@ -776,16 +926,21 @@ command line.
 
 If -A is selected each file is listed individually.  Suited for
 filtering, troubleshooting or exploration.
+
+If -f <filename> provide a summary of what information is found within.
 "))
     (if shorten-years (set! display-cprt-yrs min-max-yr-lst))
-    (if dir-provided
-	(cond
-	 (all (all-output dir-provided ex-list))
-	 (list-files #t)
-	 (#t (begin
-	       (display (string-concatenate (list "Scanning: " dir-provided "\n")))
-	       (default-output dir-provided ex-list))))
-	(display "Must provide the directory to read using the -d option. Or get --help.\n")
-	)
+    (if filename
+	(analyze-single-file filename)
+	(if dir-provided
+	    (cond
+	     (all (all-output dir-provided ex-list))
+	     (list-files #t)
+	     (#t (begin
+		   (display (string-concatenate (list "Scanning: " dir-provided "\n")))
+		   (default-output dir-provided ex-list))))
+	    (display "Must provide the directory to read using the -d option. Or get --help.\n")
+	    ))
+    (display diagnostic-str)
     )
   )
